import numpy as np
import cv2
from tacv.visual import draw_points


def is_out_side_image(x, y, w, h):
    """
    Check 2D location is outside image of shape (w,h)
    :param x:
    :param y:
    :param w:
    :param h:
    :return: True if outside image, otherwise returns False
    """
    if x < 0 or x >= w or y < 0 or y >= h:
        return True
    return False


def rotation_of_line(p1, p2):
    delta_y = p2[1] - p1[1]
    delta_x = p2[0] - p1[0]
    return np.arctan2(delta_x, delta_y)


def rotation_matrix(rotation):
    cos = np.cos(rotation)
    sin = np.sin(rotation)

    return np.array((
        [cos, -sin],
        [sin, cos])
    )


def rotate_points_around_another(p, other, rot):
    '''
    :param p: N x 2, 1st col = Xs, 2nd col = Ys
    :param other: another point. np array of 2.1
    :return: rotation in radian
    '''
    other_numpy = np.array(other).reshape(2, 1)
    point_in_other_coor_sys = np.array(p).T - other_numpy
    rot_mat = rotation_matrix(rot)
    rotated_points = rot_mat @ point_in_other_coor_sys
    rotated_points += other_numpy
    return rotated_points.T


def get_perspective_transform_matrix(src, det):
    matrix = cv2.getPerspectiveTransform(src, det)
    return matrix


def warp_image_with_perspective_transform(image, mat, dsize=(720, 1080), inplace=False):
    if inplace:
        cv2.warpPerspective(image, mat)
        return image
    else:
        new_image = cv2.warpPerspective(image.copy(), mat, dsize=dsize)
        return new_image


def get_contours(image):
    canny1 = 255 - cv2.Canny(image, threshold1=50, threshold2=200)
    # canny2 = cv2.Canny(image, threshold1=100, threshold2=200)
    # sobel = cv2.Sobel(image,ddepth=cv2.CV_8U,dy=1,dx=1,ksize=5)
    # cv2.imshow("canny1",canny1)
    # cv2.imshow("canny2", canny2)
    # cv2.imshow("sobel", sobel)
    # cv2.waitKey(0)

    image = cv2.cvtColor(image.copy(), cv2.COLOR_BGR2GRAY)
    ret, thresh = cv2.threshold(image, 127, 255, 0)
    thresh = np.where(thresh<canny1,thresh,canny1)
    contours, hierarchy = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    return contours


def get_top_k_areas(contours, k=3, take_max=True):
    contours.sort(key=lambda x: cv2.contourArea(x), reverse=take_max == True)
    return contours[:k]


def detect_corners_and_do_perspective_transform(image):
    contours = get_contours(image)
    ## only draw contour that have big areas
    imx = image.shape[0]
    imy = image.shape[1]
    image_area = imx * imy
    lp_area = image_area / 10

    ## Get only rectangles given exceeding area
    contours = get_top_k_areas(list(contours), k=5)
    approx_polygons = [cv2.approxPolyDP(cnt, 0.01 * cv2.arcLength(cnt, True), True) for cnt in contours]

    for approx, contour in zip(approx_polygons, contours):
        if len(approx) == 4 and cv2.contourArea(contour) > lp_area:
            tmp_img = image.copy()
            cv2.drawContours(tmp_img, [approx], 0, (0, 255, 255), 6)
            cv2.imshow("contour", tmp_img)
            cv2.waitKey(0)



from typing import Optional

import torch
from pytorch_lightning import LightningModule
from torch import Tensor
from torch.nn import Sequential, Conv2d, LeakyReLU, Sigmoid, ReLU
import numpy as np
from torch.nn.functional import binary_cross_entropy, mse_loss
from torch.optim import Optimizer
from torch.optim.lr_scheduler import MultiStepLR

from .losses import centerness_loss
from .utils.data_utils import gaussian_radius, draw_umich_gaussian


class CenterNet(LightningModule):
    def __init__(self, backbone, num_classes, head_conv_channel=64, max_object=64, input_shape=(480, 640),
                 train_config=None):
        super(CenterNet, self).__init__()
        self.backbone = backbone
        self.num_classes = num_classes
        self.class_head = self.init_class_head(head_conv_channel)
        self.center_off_head = self.init_center_offset_head(head_conv_channel)
        self.regress_head = self.init_regression_head(head_conv_channel)
        self.max_object = max_object
        self.input_shape = input_shape
        self.down_ratio_x = 4
        self.down_ratio_y = 4
        self.calculate_hm_shape()
        self.train_config = train_config
        self.loss_hm_reg_offset_weights = [1, 1, 0.1]
        if train_config is not None:
            self.loss_hm_reg_offset_weights = train_config["loss_hm_reg_offset_weights"]

    def calculate_hm_shape(self):
        mock_input = torch.rand(1, 3, self.input_shape[0], self.input_shape[1])
        output = self(mock_input)
        hm = output["hm"]
        h, w = hm.shape[-2:]
        self.down_ratio_x = self.input_shape[1] / w
        self.down_ratio_y = self.input_shape[0] / h

    def init_class_head(self, head_conv_channel):
        return Sequential(
            Conv2d(64, head_conv_channel,
                   kernel_size=3, padding=1, bias=True),
            LeakyReLU(inplace=True),
            Conv2d(head_conv_channel, self.num_classes,
                   kernel_size=1, stride=1,
                   padding=0, bias=True),
            Sigmoid()
        )

    def init_regression_head(self, head_conv_channel):
        return Sequential(
            Conv2d(64, head_conv_channel,
                   kernel_size=3, padding=1, bias=True),
            LeakyReLU(inplace=True),
            Conv2d(head_conv_channel, 2,
                   kernel_size=1, stride=1,
                   padding=0, bias=True),
            ReLU(inplace=True)
        )

    def init_center_offset_head(self, head_conv_channel):
        return Sequential(
            Conv2d(64, head_conv_channel,
                   kernel_size=3, padding=1, bias=True),
            LeakyReLU(inplace=True),
            Conv2d(head_conv_channel, 2,
                   kernel_size=1, stride=1,
                   padding=0, bias=True),
            Sigmoid()
        )

    def forward(self, x):
        feature = self.backbone(x)
        heatmap = self.class_head(feature)
        offset = self.center_off_head(feature)
        regression = self.regress_head(feature)
        return {
            "hm": heatmap,
            "offset": offset,
            "reg": regression
        }

    def training_step(self, batch, batch_idx):
        image = batch["image"]  # N x 3 x H X W
        annos = batch["annos"]  # N x max_obj x 5
        masks = batch["masks"]  # N x max_obj
        masks = masks.unsqueeze(dim=-1)
        model_output = self(image)
        batch_size = image.shape[0]
        num_objs = torch.nonzero(masks).__len__()
        gt_hm, gt_reg = self.create_gt_heatmap(annos, masks, batch_size)
        gt_offset = gt_reg[:, :, :2]
        gt_wh = gt_reg[:, :, 2:]
        pred_hm, pred_reg = self.gather_output(annos, masks, batch_size, model_output)
        pred_offset = pred_reg[:, :, :2]
        pred_wh = pred_reg[:, :, 2:]
        loss_hm = centerness_loss(gt_hm, pred_hm) * self.loss_hm_reg_offset_weights[0]
        loss_offset = binary_cross_entropy(pred_offset, gt_offset, reduction="none") * masks / (num_objs + 0.0001)
        loss_wh = mse_loss(pred_wh, gt_wh, reduction="none") * masks / (num_objs + 0.0001)
        loss_offset = loss_offset.sum() * self.loss_hm_reg_offset_weights[1]
        loss_wh = loss_wh.sum() * self.loss_hm_reg_offset_weights[2]
        total = loss_hm + loss_offset + loss_wh
        print("Losses:", f"Train Loss hm: {loss_hm}, Loss offset: {loss_offset}, Loss wh: {loss_wh}")
        return total

    @torch.no_grad()
    def validation_step(self, val_batch, batch_idx):
        image = val_batch["image"]  # N x 3 x H X W
        annos = val_batch["annos"]  # N x max_obj x 4
        masks = val_batch["masks"]  # N x max_obj
        masks = masks.unsqueeze(dim=-1)
        model_output = self(image)
        batch_size = image.shape[0]
        num_objs = torch.nonzero(masks).__len__()
        gt_hm, gt_reg = self.create_gt_heatmap(annos, masks, batch_size)
        gt_offset = gt_reg[:, :, :2]
        gt_wh = gt_reg[:, :, 2:]
        pred_hm, pred_reg = self.gather_output(annos, masks, batch_size, model_output)
        pred_offset = pred_reg[:, :, :2]
        pred_wh = pred_reg[:, :, 2:]
        loss_hm = centerness_loss(gt_hm, pred_hm) * self.loss_hm_reg_offset_weights[0]
        loss_offset = binary_cross_entropy(pred_offset, gt_offset, reduction="none") * masks / (num_objs + 0.0001)
        loss_wh = mse_loss(pred_wh, gt_wh, reduction="none") * masks / (num_objs + 0.0001)
        loss_offset = loss_offset.sum() * self.loss_hm_reg_offset_weights[1]
        loss_wh = loss_wh.sum() * self.loss_hm_reg_offset_weights[2]
        total = loss_hm + loss_offset + loss_wh
        print("Val:", f"Train Loss hm: {loss_hm}, Loss offset: {loss_offset}, Loss wh: {loss_wh}")
        self.log("val_loss", total)

    def backward(self, loss: Tensor, optimizer: Optional[Optimizer], optimizer_idx: Optional[int], *args,
                 **kwargs) -> None:
        loss.backward()

    def gather_output(self, annos, masks, batch_size, output):
        """
        :param annos:
        :param masks:
        :param batch_size:
        :param output: model output
        :return:
        """
        hm = output["hm"]
        reg = torch.zeros((batch_size, self.max_object, 4), device=self.device)  # 2 for offset and 2 for wh

        model_reg_output = output["reg"]
        model_offset_output = output["offset"]

        for image_index, (image_annos, obj_masks) in enumerate(zip(annos, masks)):
            for obj_index, (value, anno) in enumerate(zip(obj_masks, image_annos)):
                if value.item() is True:
                    x, y, w, h, class_id = anno
                    x, y, w, h = x.item(), y.item(), w.item(), h.item()
                    down_y = y / self.down_ratio_y
                    down_x = x / self.down_ratio_x

                    down_y_int = int(down_y)
                    down_x_int = int(down_x)
                    output_offset_x = model_offset_output[image_index, 0, down_y_int, down_x_int]
                    output_offset_y = model_offset_output[image_index, 1, down_y_int, down_x_int]
                    output_w = model_reg_output[image_index, 0, down_y_int, down_x_int]
                    output_h = model_reg_output[image_index, 1, down_y_int, down_x_int]

                    reg[image_index, obj_index, 0] = output_offset_x
                    reg[image_index, obj_index, 1] = output_offset_y
                    reg[image_index, obj_index, 2] = output_w
                    reg[image_index, obj_index, 3] = output_h
                else:
                    continue
        return hm, reg

    def create_gt_heatmap(self, annos, masks, batch_size):
        hm_x = self.input_shape[1] / self.down_ratio_x
        hm_y = self.input_shape[0] / self.down_ratio_y
        hm = np.zeros((batch_size, self.num_classes, int(hm_y), int(hm_x)))
        reg = torch.zeros((batch_size, self.max_object, 4), device=self.device)  # 2 for offset and 2 for wh
        for index, (image_annos, obj_masks) in enumerate(zip(annos, masks)):
            for obj_index, (value, anno) in enumerate(zip(obj_masks, image_annos)):
                if value.item():
                    x, y, w, h, class_id = anno
                    x, y, w, h = x.item(), y.item(), w.item(), h.item()
                    class_id = int(class_id)

                    down_y = y / self.down_ratio_y
                    down_x = x / self.down_ratio_x

                    down_y_int = int(down_y)
                    down_x_int = int(down_x)
                    # build hm
                    radius = gaussian_radius((h, w))
                    radius = max(0, int(radius))
                    hm[index, class_id, :, :] = draw_umich_gaussian(hm[index, class_id], (down_x_int, down_y_int),
                                                                    radius)
                    # build regression gt
                    offset_y = down_y - down_y_int
                    offset_x = down_x - down_x_int

                    reg[index, obj_index, 0] = offset_x
                    reg[index, obj_index, 1] = offset_y
                    reg[index, obj_index, 2] = w
                    reg[index, obj_index, 3] = h
                else:
                    continue
        hm = torch.as_tensor(hm, device=self.device)
        return hm, reg

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.train_config["learning_rate"],
                                     weight_decay=self.train_config["weight_decay"])
        lr_scheduler = MultiStepLR(optimizer, milestones=self.train_config["lr_decay_milestones"],
                                   gamma=self.train_config["lr_decay_gamma"])
        return [optimizer], [lr_scheduler]



from typing import Union
import torch

def load_weights(model, source: Union[str,torch.nn.Module], name_matching_required=False, shape_matching_required=False, location="cpu", name_must_match: list = []):
    if isinstance(source,str):
        state_dict = torch.load(source, map_location=location)
    else:
        state_dict = source
    for name, param in model.named_parameters():
        if name not in state_dict:
            if name_matching_required:
                return False
            else:
                continue
        if param.shape != state_dict[name].shape:
            if shape_matching_required:
                return False
            else:
                continue
        if len(name_must_match)!=0:
            has_the_layers = False
            for specified_name in name_must_match:
                if specified_name in name:
                    has_the_layers = True
                    break
            if not has_the_layers:
                continue
        param.data.copy_(state_dict[name])
    return True



import sys
from ..video import video2images, images2video


def cli_images2video():
    args = sys.argv[1:]
    images2video(*args)


def cli_video2images():
    args = sys.argv[1:]
    video2images(*args)



from typing import Union

import cv2
import numpy as np

def draw_points(image,points: Union[list,tuple,np.ndarray],circular=True,color=(255,0,0),thickness=2):
    '''
    :param image:
    :param points: Must be in shape of N x 2, where N is the number of points
    :param circular:
    :param color:
    :param thickness:
    :return:
    '''
    points = [(int(point[0]),int(point[1])) for point in points]
    N = len(points)
    if not circular:
        N = N - 1
    for i in range(N):
        cv2.line(image,points[i],points[(i+1)%N],color,thickness)


import numpy as np


def transpose_image(image, channel_first=False):
    def transpose_single_image(image, channel_first):
        if image.shape[0] <= 3 and channel_first == False:  # channel first
            image = image.transpose((1, 2, 0))
        if image.shape[0] > 3 and channel_first == True:
            image = image.transpose((2, 0, 1))
        return image

    def transpose_batch_of_images(images, channel_first):
        if image.shape[1] <= 3 and channel_first == False:  # channel first
            images = image.transpose((0, 2, 3, 1))
        if image.shape[1] > 3 and channel_first == True:
            images = image.transpose((0, 3, 1, 2))
        return images

    shape = image.shape
    if len(shape) == 3:
        image = transpose_single_image(image, channel_first)
    elif len(shape) == 4:
        image = transpose_batch_of_images(image, channel_first)
    return image


def gaussian_radius(det_size, min_overlap=0.7):
    height, width = det_size

    a1 = 1
    b1 = (height + width)
    c1 = width * height * (1 - min_overlap) / (1 + min_overlap)
    sq1 = np.sqrt(b1 ** 2 - 4 * a1 * c1)
    r1 = (b1 + sq1) / 2

    a2 = 4
    b2 = 2 * (height + width)
    c2 = (1 - min_overlap) * width * height
    sq2 = np.sqrt(b2 ** 2 - 4 * a2 * c2)
    r2 = (b2 + sq2) / 2

    a3 = 4 * min_overlap
    b3 = -2 * min_overlap * (height + width)
    c3 = (min_overlap - 1) * width * height
    sq3 = np.sqrt(b3 ** 2 - 4 * a3 * c3)
    r3 = (b3 + sq3) / 2
    return min(r1, r2, r3)


def gaussian2D(shape, sigma=1):
    m, n = [(ss - 1.) / 2. for ss in shape]
    y, x = np.ogrid[-m:m + 1, -n:n + 1]

    h = np.exp(-(x * x + y * y) / (2 * sigma * sigma))
    h[h < np.finfo(h.dtype).eps * h.max()] = 0
    return h


def draw_umich_gaussian(heatmap, center, radius, k=1):
    diameter = 2 * radius + 1
    gaussian = gaussian2D((diameter, diameter), sigma=diameter / 6)

    x, y = int(center[0]), int(center[1])

    height, width = heatmap.shape[0:2]

    left, right = min(x, radius), min(width - x, radius + 1)
    top, bottom = min(y, radius), min(height - y, radius + 1)

    masked_heatmap = heatmap[y - top:y + bottom, x - left:x + right]
    masked_gaussian = gaussian[radius - top:radius + bottom, radius - left:radius + right]
    if min(masked_gaussian.shape) > 0 and min(masked_heatmap.shape) > 0:  # TODO debug
        np.maximum(masked_heatmap, masked_gaussian * k, out=masked_heatmap)
    return heatmap
